@@ -1,4 +1,5 @@
 import asyncio
+import json
 import logging
 import os
 import signal
@@ -7,7 +8,7 @@ import traceback
 from asyncio import AbstractEventLoop
 from datetime import datetime
 from threading import Thread
-from typing import Optional, List, Dict, Set
+from typing import Optional, List, Dict, Set, Awaitable, Callable, Tuple
 
 from dacite import from_dict
 from ffmpy import FFmpeg, FFRuntimeError
@@ -19,8 +20,10 @@ from TikTokLive.client import config
 from TikTokLive.client.httpx import TikTokHTTPClient
 from TikTokLive.proto.tiktok_schema_pb2 import WebcastWebsocketAck
 from TikTokLive.proto.utilities import deserialize_websocket_message, serialize_message
-from TikTokLive.types import AlreadyConnecting, AlreadyConnected, LiveNotFound, FailedConnection, ExtendedGift, InvalidSessionId, ChatMessageSendFailure, ChatMessageRepeat, FailedFetchRoomInfo, FailedFetchGifts, \
-    FailedRoomPolling, FFmpegWrapper, AlreadyDownloadingStream, DownloadProcessNotFound, NotDownloadingStream, InitialCursorMissing
+from TikTokLive.types import AlreadyConnecting, AlreadyConnected, LiveNotFound, FailedConnection, ExtendedGift, \
+    FailedFetchRoomInfo, FailedFetchGifts, \
+    FailedRoomPolling, FFmpegWrapper, AlreadyDownloadingStream, DownloadProcessNotFound, NotDownloadingStream, \
+    InitialCursorMissing, VideoQuality, InvalidSessionId, ChatMessageRepeat, ChatMessageSendFailure
 from TikTokLive.utils import validate_and_normalize_unique_id, get_room_id_from_main_page_html
 
 
@@ -35,7 +38,8 @@ class BaseClient(AsyncIOEventEmitter):
             unique_id: str,
             loop: Optional[AbstractEventLoop] = None,
             client_params: Optional[dict] = None,
-            headers: Optional[dict] = None,
+            request_headers: Optional[dict] = None,
+            websocket_headers: Optional[dict] = None,
             timeout_ms: Optional[int] = None,
             ping_interval_ms: int = 1000,
             process_initial_data: bool = True,
@@ -53,7 +57,7 @@ class BaseClient(AsyncIOEventEmitter):
         :param unique_id: The unique id of the creator to connect to
         :param loop: Optionally supply your own asyncio loop
         :param client_params: Additional client parameters to include when making requests to the Webcast API
-        :param headers: Additional headers to include when making requests to the Webcast API
+        :param request_headers: Additional headers to include when making requests to the Webcast API
         :param timeout_ms: The timeout (in ms) for requests made to the Webcast API
         :param ping_interval_ms: The interval between requests made to the Webcast API for both Websockets and Long Polling
         :param process_initial_data: Whether to process the initial data (including cached chats)
@@ -65,51 +69,60 @@ class BaseClient(AsyncIOEventEmitter):
         :param websocket_enabled: Whether to use websockets or rely on purely long polling
         :param sign_api_key: Parameter to increase the amount of connections allowed to be made per minute via a Sign Server API key. If you need this, contact the project maintainer.
         """
-        AsyncIOEventEmitter.__init__(self)
 
-        # Get Event Loop
-        if isinstance(loop, AbstractEventLoop):
-            self.loop: AbstractEventLoop = loop
-        else:
-            try:
-                self.loop: AbstractEventLoop = asyncio.get_running_loop()
-            except RuntimeError:
-                self.loop: AbstractEventLoop = asyncio.new_event_loop()
+        AsyncIOEventEmitter.__init__(self)
+        self.loop: AbstractEventLoop = self.__get_event_loop(loop=loop)
 
         # Private Attributes
         self.__unique_id: str = validate_and_normalize_unique_id(unique_id)
-        self.__discard_extra_events: Optional[bool] = None
         self.__room_info: Optional[dict] = None
         self.__available_gifts: Dict[int, ExtendedGift] = dict()
-        self.__room_id: Optional[str] = None
+        self.__room_id: Optional[int] = None
         self._viewer_count: Optional[int] = None
         self.__connecting: bool = False
         self.__connected: bool = False
         self.__session_id: Optional[str] = None
         self.__is_ws_upgrade_done: bool = False
         self.__websocket_enabled: bool = websocket_enabled
+        self.__websocket_headers: Optional[dict] = websocket_headers if isinstance(websocket_headers, dict) else dict()
 
         # Change Language
         config.DEFAULT_CLIENT_PARAMS["app_language"] = lang
         config.DEFAULT_CLIENT_PARAMS["webcast_language"] = lang
 
         # Protected Attributes
+        self._ping_interval_ms: int = ping_interval_ms
+        self._process_initial_data: bool = process_initial_data
+        self._enable_extended_gift_info: bool = enable_extended_gift_info
+        self._fetch_room_info_on_connect: bool = fetch_room_info_on_connect
+        self._download: Optional[FFmpegWrapper] = None
+        self._first_connect: bool = True
         self._http: TikTokHTTPClient = TikTokHTTPClient(
-            headers=headers if headers is not None else dict(),
+            headers=request_headers if request_headers is not None else dict(),
             timeout_ms=timeout_ms,
             proxies=proxies,
             trust_env=trust_env,
             params={**config.DEFAULT_CLIENT_PARAMS, **(client_params if isinstance(client_params, dict) else dict())},
             sign_api_key=sign_api_key
         )
-        self._ping_interval_ms: int = ping_interval_ms
-        self._process_initial_data: bool = process_initial_data
-        self._enable_extended_gift_info: bool = enable_extended_gift_info
-        self._fetch_room_info_on_connect: bool = fetch_room_info_on_connect
-        self._download: Optional[FFmpegWrapper] = None
 
-        # Listeners
-        self.add_listener("websocket", self._handle_webcast_messages)
+    @classmethod
+    def __get_event_loop(cls, loop: Optional[AbstractEventLoop]) -> AbstractEventLoop:
+        """
+        Get event loop for constructor
+
+        :param loop: Loop object to validate
+        :return: A valid event loop
+
+        """
+
+        if isinstance(loop, AbstractEventLoop):
+            return loop
+        else:
+            try:
+                return asyncio.get_running_loop()
+            except RuntimeError:
+                return asyncio.new_event_loop()
 
     async def _on_error(self, original: Exception, append: Optional[Exception]) -> None:
         """
@@ -123,7 +136,7 @@ class BaseClient(AsyncIOEventEmitter):
 
         raise NotImplementedError()
 
-    async def __fetch_room_id(self) -> Optional[str]:
+    async def __fetch_room_id(self) -> Optional[int]:
         """
         Fetch room ID of a given user
 
@@ -134,8 +147,8 @@ class BaseClient(AsyncIOEventEmitter):
 
         try:
             html: str = await self._http.get_livestream_page_html(self.__unique_id)
-            self.__room_id = get_room_id_from_main_page_html(html)
-            self._http.params["room_id"] = self.__room_id
+            self.__room_id = int(get_room_id_from_main_page_html(html))
+            self._http.params["room_id"] = str(self.__room_id)
             return self.__room_id
         except Exception as ex:
             await self._on_error(ex, FailedFetchRoomInfo("Failed to fetch room id from Webcast, see stacktrace for more info."))
@@ -201,7 +214,7 @@ class BaseClient(AsyncIOEventEmitter):
 
             await asyncio.sleep(polling_interval)
 
-    async def __fetch_room_data(self, is_initial: bool = False) -> None:
+    async def __fetch_room_data(self, is_initial: bool = False, first_connect: bool = True) -> None:
         """
         Fetch room data from the Webcast API and deserialize it
 
@@ -211,7 +224,13 @@ class BaseClient(AsyncIOEventEmitter):
         """
 
         # Fetch from polling api
-        webcast_response = await self._http.get_deserialized_object_from_webcast_api("im/fetch/", self._http.params, "WebcastResponse", is_initial)
+        webcast_response = await (
+            self._http.get_deserialized_object_from_signing_api("webcast/fetch/", self._http.params, "WebcastResponse")
+            if is_initial else
+            self._http.get_deserialized_object_from_signing_api("im/fetch/", self._http.params, "WebcastResponse")
+        )
+
+        # Cursor
         _last_cursor, _next_cursor = self._http.params["cursor"], webcast_response.get("cursor")
         self._http.params["cursor"] = _last_cursor if _next_cursor == "0" else _next_cursor
 
@@ -231,7 +250,8 @@ class BaseClient(AsyncIOEventEmitter):
             if not self._process_initial_data:
                 return
 
-        await self._handle_webcast_messages(webcast_response)
+        if first_connect:
+            await self._handle_webcast_messages(webcast_response)
 
     async def __try_websocket_upgrade(self, webcast_response) -> None:
         """
@@ -248,7 +268,8 @@ class BaseClient(AsyncIOEventEmitter):
         )
 
         headers: dict = {
-            "Cookie": " ".join(f"{k}={v};" for k, v in self._http.client.cookies.items())
+            **{"Cookie": " ".join(f"{k}={v};" for k, v in self._http.client.cookies.items())},
+            **self.__websocket_headers
         }
 
         try:
@@ -347,7 +368,7 @@ class BaseClient(AsyncIOEventEmitter):
 
         raise NotImplementedError
 
-    async def _connect(self, session_id: str = None) -> str:
+    async def _connect(self, session_id: str = None) -> int:
         """
         Connect to the WebcastWebsocket API
 
@@ -380,7 +401,7 @@ class BaseClient(AsyncIOEventEmitter):
                 await self.__fetch_available_gifts()
 
             # Make initial request to Webcast Messaging
-            await self.__fetch_room_data(True)
+            await self.__fetch_room_data(True, first_connect=self._first_connect)
             self.__connected = True
 
             # If the websocket was not connected for whatever reason
@@ -396,6 +417,7 @@ class BaseClient(AsyncIOEventEmitter):
                         + "Long polling is not available: Try adding a sessionid as an argument in start() or run()"
                     )
 
+            self._first_connect = False
             return self.__room_id
 
         except Exception as ex:
@@ -429,7 +451,7 @@ class BaseClient(AsyncIOEventEmitter):
         self._http.params["cursor"]: str = ""
 
         if webcast_closed:
-            logging.error("Connection was lost to the Webcast Websocket Server. Restart the client connection to continue.")
+            logging.error("Connection was lost to the Webcast Websocket Server. Use await client.reconnect() to reconnect.")
 
     def stop(self) -> None:
         """
@@ -440,10 +462,10 @@ class BaseClient(AsyncIOEventEmitter):
         """
 
         if self.__connected:
-            self._disconnect()
+            self._disconnect(webcast_closed=False)
             return
 
-    async def start(self, session_id: Optional[str] = None) -> Optional[str]:
+    async def start(self, session_id: Optional[str] = None) -> Optional[int]:
         """
         Start the client without blocking the main thread
 
@@ -474,42 +496,8 @@ class BaseClient(AsyncIOEventEmitter):
         """
 
         if session_id:
-            self.__session_id = session_id
             self._http.client.cookies.set("sessionid", session_id)
 
-    async def send_message(self, text: str, session_id: Optional[str] = None) -> Optional[str]:
-        """
-        Send a message to the TikTok Live Chat
-
-        :param text: The message you want to send to the chat
-        :param session_id: The Session ID (If you've already supplied one, you don't need to)
-        :return: The response from the webcast API
-
-        """
-
-        self.__set_session_id(session_id)
-
-        if not self.__session_id:
-            raise InvalidSessionId("Missing Session ID. Please provide your current Session ID to use this feature.")
-
-        params: dict = {**self._http.params, "content": text}
-        response: dict = await self._http.post_json_to_webcast_api("room/chat/", params, None, sign_url=False)
-        status_code: Optional[int] = response.get("status_code")
-        data: Optional[dict] = response.get("data")
-
-        if status_code == 0:
-            return data
-
-        try:
-            raise {
-                20003: InvalidSessionId("Your Session ID has expired. Please provide a new one"),
-                50007: ChatMessageRepeat("You cannot send repeated chat messages!")
-            }.get(
-                status_code, ChatMessageSendFailure(f"TikTok responded with status code {status_code}: {data.get('message')}")
-            )
-        except Exception as ex:
-            await self._on_error(ex, None)
-
     async def retrieve_room_info(self) -> Optional[dict]:
         """
         Method to retrieve room information
@@ -539,6 +527,7 @@ class BaseClient(AsyncIOEventEmitter):
             self,
             path: str,
             duration: Optional[int] = None,
+            quality: Optional[VideoQuality] = None,
             verbose: bool = True,
             loglevel: str = "error",
             global_options: Set[str] = set(),
@@ -554,6 +543,7 @@ class BaseClient(AsyncIOEventEmitter):
         :param global_options: Pass custom params to FFmpeg global options
         :param path: The path to download the livestream video to
         :param duration: If duration is None or less than 1, download will go forever
+        :param quality: If quality is None, download quality will auto
         :param verbose: Whether to log info about the download in console
 
         :return: None
@@ -570,6 +560,15 @@ class BaseClient(AsyncIOEventEmitter):
         if duration is not None and duration >= 1:
             runtime = f"-t {duration}"
 
+        # Set a quality
+        url: dict = json.loads(self.room_info['stream_url']['live_core_sdk_data']['pull_data']['stream_data'])
+        quality = quality if isinstance(quality, VideoQuality) else VideoQuality.ORIGIN
+
+        # Set the URL based on selected quality
+        url_param: str = url['data'][quality.value]['main']['hls']
+        if len(url_param.strip()) == 0:
+            url_param: str = url['data'][quality.value]['main']['flv']
+
         # Function Running
         def spool():
             try:
@@ -583,7 +582,7 @@ class BaseClient(AsyncIOEventEmitter):
         # Create an FFmpeg wrapper
         self._download = FFmpegWrapper(
             ffmpeg=FFmpeg(
-                inputs={**{self.room_info["stream_url"]["hls_pull_url"]: None}, **inputs},
+                inputs={**{url_param: None}, **inputs},
                 outputs={**{path: runtime}, **outputs},
                 global_options={"-y", f"-loglevel {loglevel}"}.union(global_options)
             ),
@@ -599,7 +598,53 @@ class BaseClient(AsyncIOEventEmitter):
 
         # Give info about the started download
         if self._download.verbose:
-            logging.warning(f"Started the download to path \"{path}\" for duration \"{'infinite' if runtime is None else duration} seconds\" on user @{self.unique_id}")
+            logging.warning(f"Started the download to path \"{path}\" for duration \"{'infinite' if runtime is None else duration} seconds\" on user @{self.unique_id} with \"{quality.name}\" video quality")
+
+    async def send_message(self, text: str, sign_url_fn: Callable[[str, str], Awaitable[Tuple[str, Dict[str, str]]]], session_id: Optional[str] = None) -> Optional[str]:
+        """
+        Send a message to the TikTok Live Chat.
+
+        Note that you will have to implement a sign_url_fn coroutine function
+        that will take two parameters, a STRING for the unsigned URL and a STRING for the sessionid.
+
+        This function MUST return a signed URL STRING and headers DICTIONARY to be used in the request.
+
+        :param sign_url_fn: Function that takes in two *args, URL (str) and SessionID (str) and returns two args, Signed URL (str) and Headers (dict)
+        :param text: The message you want to send to the chat
+        :param session_id: The Session ID (If you've already supplied one, you don't need to)
+        :return: The response from the webcast API
+        """
+
+        # Set session ID if given
+        self.__set_session_id(session_id)
+
+        # Check a session ID is provided
+        if not self._http.client.cookies.get("sessionid"):
+            raise InvalidSessionId("Missing Session ID. Please provide your current Session ID to use this feature.")
+
+        # Build a URL
+        params: dict = {**self._http.params, "content": text}
+        raw_url: str = self._http.update_url((config.TIKTOK_URL_WEBCAST + "room/chat/"), params if params else dict())
+
+        # Sign the URL
+        signed_url, headers = await sign_url_fn(raw_url, self._http.client.cookies.get("sessionid"))
+        response: dict = await self._http.post_json_to_url(signed_url, headers, None)
+
+        status_code: Optional[int] = response.get("status_code")
+        data: Optional[dict] = response.get("data")
+
+        if status_code == 0:
+            return data
+
+        try:
+            raise {
+                20003: InvalidSessionId("Your Session ID has expired. Please provide a new one"),
+                50007: ChatMessageRepeat("You cannot send repeated chat messages!")
+            }.get(
+                status_code, ChatMessageSendFailure(f"TikTok responded with status code {status_code}: {data.get('message')}")
+            )
+        except Exception as ex:
+            await self._on_error(ex, None)
 
     def stop_download(self) -> None:
         """
@@ -629,35 +674,29 @@ class BaseClient(AsyncIOEventEmitter):
                 f"\"{int(datetime.utcnow().timestamp()) - self._download.started_at} seconds\" of downloading"
             )
 
-    async def set_proxies(self, proxies: Optional[Dict[str, str]]) -> None:
+    @property
+    def proxies(self) -> Optional[Dict[str, str]]:
+        """
+        Get the current proxies being used in HTTP requests
+
+        """
+        return self._http.proxies
+
+    @proxies.setter
+    def proxies(self, proxies: Optional[Dict[str, str]]) -> None:
         """
         Set the proxies to be used by the HTTP client (Not Websockets)
 
         :param proxies: The proxies to use in HTTP requests
-        :return: None
 
         """
-
         self._http.proxies = proxies
 
-    @property
-    async def proxies(self) -> Optional[Dict[str, str]]:
-        """
-        Get the current proxies being used in HTTP requests
-
-        :return: The current proxies in use
-
-        """
-
-        return self._http.proxies
-
     @property
     def viewer_count(self) -> Optional[int]:
         """
         Return viewer count of user
 
-        :return: Viewer count
-
         """
         return self._viewer_count
 
@@ -666,8 +705,6 @@ class BaseClient(AsyncIOEventEmitter):
         """
         Room ID if the connection was successful
 
-        :return: Room's ID
-
         """
         return self.__room_id
 
@@ -676,10 +713,7 @@ class BaseClient(AsyncIOEventEmitter):
         """
         Room info dict if the connection was successful
 
-        :return: Room Info Dict
-
         """
-
         return self.__room_info
 
     @property
@@ -687,10 +721,7 @@ class BaseClient(AsyncIOEventEmitter):
         """
         Unique ID of the streamer
 
-        :return: Their unique ID
-
         """
-
         return self.__unique_id
 
     @property
@@ -698,8 +729,6 @@ class BaseClient(AsyncIOEventEmitter):
         """
         Whether the client is connected
 
-        :return: Result
-
         """
 
         return self.__connected
@@ -709,8 +738,6 @@ class BaseClient(AsyncIOEventEmitter):
         """
         Available gift information for live room
 
-        :return: Gift info
-
         """
 
         return self.__available_gifts
